/*
 *	Linux Process Freezer
 *
 *	Copyright (c) 1998 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

/*
 *  WARNING: Very dirty. Very experimental. Very PeeCee-centric.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/signal.h>
#include <asm/page.h>

static int pid;

static void __attribute__((noreturn))
die(char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fprintf(stderr, "freeze: ");
  vfprintf(stderr, msg, args);
  fputc('\n', stderr);
  exit(1);
}

static void
chld_handler(int __attribute__((unused)) unused)
{
  puts("SIGCHLD");
}

static void
wr(int fd, void *buf, int size)
{
  if (write(fd, buf, size) != size)
    die("write: %m");
}

static inline void
ls(int fd, unsigned long offs)
{
  lseek(fd, offs, SEEK_SET);
}

static void
pr(int fd, char *msg, ...)
{
  va_list args;
  char buf[2048];
  int len;

  va_start(args, msg);
  vsprintf(buf, msg, args);
  len = strlen(buf) + 1;
  wr(fd, &len, sizeof(len));
  wr(fd, buf, len);
  va_end(args);
}

static void
save_map(int fd, int mem, char *attrs, unsigned long base, unsigned long size, unsigned long offs, char *file)
{
  char page[PAGE_SIZE], page2[PAGE_SIZE];
  int f2;

  printf("Saving %s region @%lx, len %lx, offset %lx in %s\n", attrs, base, size, offs, file ? : "ANON");
  if (size & (PAGE_SIZE - 1))
    die("Size not page-aligned");
  if (file)
    {
      pr(fd, "M %.3s %lx %lx %lx %s", attrs, base, size, offs, file);
      if (attrs[1] == '-')
	return;
      if ((f2 = open(file, O_RDONLY)) < 0)
	die("Mapping %s: %m", file);
    }
  else
    {
      if (attrs[1] == '-')
	die("Non-writeable anonymous mappings are not supported");
      pr(fd, "M %.3s %lx %lx", attrs, base, size);
      f2 = -1;
      bzero(page, sizeof(page));
    }
  ls(mem, base);

  while (size > 0)
    {
      int t;
      if (f2 >= 0 && read(f2, page, PAGE_SIZE) != PAGE_SIZE)
	die("Mapped file read: %m");
      t = read(mem, page2, PAGE_SIZE);
      if (t && t != PAGE_SIZE)
	die("Address space read: %m");
      if (!t)
	{
#if 1
	  unsigned long data;
#if 0
	  puts("WARNING");
#endif
	  if (ptrace(PTRACE_PEEKDATA, pid, base, &data) < 0)
	    die("Page at %lx is unreadable", base);
	  if (read(mem, page2, PAGE_SIZE) != PAGE_SIZE)
	    die("Error scanning page at %lx", base);
#else
	  printf("%lx: Page not mapped, skipping.\n", base); /* FIXME */
	  ls(mem, base+PAGE_SIZE);
#endif
	}
      else if (memcmp(page, page2, PAGE_SIZE))
	{
	  printf("Writing block at %lx\n", base);
	  pr(fd, "W %lx %x", base, PAGE_SIZE);
	  wr(fd, page2, PAGE_SIZE);
	}
      base += PAGE_SIZE;
      size -= PAGE_SIZE;
    }
  if (f2 >= 0)
    close(f2);
  puts("DONE");
}

static void
dump_regs(int fd)
{
  unsigned long regs[SS+1];

  if (ptrace(PTRACE_GETREGS, pid, 0, regs) < 0)
    die("PTRACE_GETREGS: %m");
  pr(fd, "R %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx",
     regs[EAX], regs[EBX], regs[ECX], regs[EDX], regs[ESI], regs[EDI],
     regs[EBP], regs[UESP], regs[EIP], regs[EFL]);
  printf("Regs: %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
     regs[EAX], regs[EBX], regs[ECX], regs[EDX], regs[ESI], regs[EDI],
     regs[EBP], regs[UESP], regs[EIP], regs[EFL]);
  /* FP registers ... FIXME!!! */
}

int
main(int argc, char **argv)
{
  sigset_t sset;
  struct sigaction sact;
  char procname[256], *p, lbuf[1024];
  FILE *f;
  int fd, mem;

  if (argc != 3)
    {
      fprintf(stderr, "Usage: freeze <PID> <file>\n");
      return 1;
    }
  pid = atol(argv[1]);
  fd = open(argv[2], O_WRONLY | O_CREAT | O_TRUNC, 0666);
  if (fd < 0)
    die("%s: %m", argv[2]);

  sigemptyset(&sset);
  sigaddset(&sset, SIGCHLD);
  if (sigprocmask(SIG_BLOCK, &sset, NULL) < 0)
    die("sigprocmask: %m");
  sigemptyset(&sact.sa_mask);
  sact.sa_handler = chld_handler;
  sact.sa_flags = 0;
  if (sigaction(SIGCHLD, &sact, NULL) < 0)
    die("sigaction: %m");

  printf("Freezing PID %d\n", pid);
  if (ptrace(PT_ATTACH, pid, 0, 0) < 0)
    die("ptrace(PT_ATTACH): %m");

  puts("Closing refrigerator door...");
  sigemptyset(&sset);
  while (sigsuspend(&sset) >= 0 || errno != EINTR)
    die("sigsuspend: %m");

  puts("Closed. Snow is about to fall...");

  dump_regs(fd);

  p = procname + sprintf(procname, "/proc/%d/", pid);
  strcpy(p, "mem");
  if ((mem = open(procname, O_RDONLY)) < 0)
    die("%s: %m", procname);
  strcpy(p, "maps");
  if (!(f = fopen(procname, "r")))
    die("%s: %m", procname);
  while (fgets(lbuf, sizeof(lbuf)-1, f))
    {
      unsigned long base, top, offs;
      char *q = strchr(lbuf, '\n');
      char *attrs, *file;
      if (!q)
	die("maps: line too long or incomplete");
      *q = 0;
      if (sscanf(lbuf, "%lx-%lx", &base, &top) != 2 || !(q = strchr(lbuf, ' ')))
	die("%s: syntax error #1 (%s)", procname, lbuf);
      attrs = ++q;
      if (q[0] != 'r' ||
	  (q[1] != 'w' && q[1] != '-') ||
	  (q[2] != 'x' && q[2] != '-') ||
	  q[3] != 'p' ||
	  q[4] != ' ')
	die("Unable to dump memory zone of type %.4s. Giving up.", q);
      q += 5;
      if (sscanf(q, "%lx", &offs) != 1 ||
	  !(q = strchr(q, ' ')))
	die("%s: syntax error #2 (%s)", procname, lbuf);
      q++;
      if (!strncmp(q, "00:00 ", 6))
	{
	  attrs[3] = 'A';
	  file = NULL;
	}
      else if (!(q = strchr(q, ' ')) || !(q = strchr(q+1, ' ')))
	die("%s: syntax error #3 (%s)", procname, lbuf);
      else
	{
	  while (*q == ' ')
	    q++;
	  file = q;
	}
      attrs[4] = 0;
      save_map(fd, mem, attrs, base, top-base, offs, file);
    }
  fclose(f);
  close(mem);

  if (ptrace(PT_DETACH, pid, 0, 0) >= 0)
    puts("Process resumed.");

  return 0;
}
